﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Management;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading.Tasks;
using Microsoft.Win32.SafeHandles;
using static ByteHandler.UnsafeNativeMethods;

namespace ByteHandler
{
    internal class LowLevelDiskAccess
    {
        /// <summary> 
        /// Returns the number of bytes that the drive sectors contain. 
        /// </summary> 
        /// <param name="drive"> 
        /// Int: The drive number to scan. 
        /// </param> 
        /// <returns> 
        /// Int: The number of bytes the sector contains. 
        /// </returns> 
        public int BytesPerSector(int drive)
        {
            int driveCounter = 0;
            try
            {
                ManagementObjectSearcher searcher =
                    new ManagementObjectSearcher("root\\CIMV2",
                    "SELECT * FROM Win32_DiskDrive");

                foreach (ManagementObject queryObj in searcher.Get())
                {
                    if (driveCounter == drive)
                    {
                        var t = queryObj["BytesPerSector"];
                        return int.Parse(t.ToString());

                    }
                    driveCounter++;
                }
            }
            catch (ManagementException)
            {
                return -1;
            }
            return 0;
        }

        /// <summary> 
        /// Returns a list of physical drive IDs 
        /// </summary> 
        /// <returns> 
        /// List<string>: Device IDs of all connected physical hard drives 
        ///  </returns> 
        public List<string> GetDriveList()
        {
            List<string> drivelist = new List<string>();

            try
            {
                ManagementObjectSearcher searcher =
                    new ManagementObjectSearcher("root\\CIMV2",
                    "SELECT * FROM Win32_DiskDrive");

                foreach (ManagementObject queryObj in searcher.Get())
                {
                    drivelist.Add(queryObj["DeviceID"].ToString());
                }
            }
            catch (ManagementException)
            {
                return null;
            }
            return drivelist;
        }

        /// <summary> 
        /// Returns the total sectors on the specified drive 
        /// </summary> 
        /// <param name="drive"> 
        /// int: The drive to be queried. 
        /// </param> 
        /// <returns> 
        /// int: Returns the total number of sectors 
        /// </returns> 
        public static int GetTotalSectors(int drive)
        {
            int driveCount = 0;
            try
            {
                ManagementObjectSearcher searcher =
                    new ManagementObjectSearcher("root\\CIMV2",
                    "SELECT * FROM Win32_DiskDrive");

                foreach (ManagementObject queryObj in searcher.Get())
                {
                    if (driveCount == drive)
                    {
                        var t = queryObj["TotalSectors"];
                        return int.Parse(t.ToString());

                    }
                    driveCount++;
                }
            }
            catch (ManagementException)
            {
                return -1;
            }
            return -1;
        }

        /// <summary> 
        /// Returns the number of Sectors per track. 
        /// </summary> 
        /// <param name="drive"> 
        /// The drive to be queried. 
        /// </param> 
        /// <returns> 
        /// int: the number of sectors per track 
        /// </returns> 
        public static int GetSectorsPerTrack(int drive)
        {
            int driveCount = 0;
            try
            {
                ManagementObjectSearcher searcher =
                    new ManagementObjectSearcher("root\\CIMV2",
                    "SELECT * FROM Win32_DiskDrive");

                foreach (ManagementObject queryObj in searcher.Get())
                {
                    if (driveCount == drive)
                    {
                        var t = queryObj["SectorsPerTrack"];
                        return int.Parse(t.ToString());
                    }
                    driveCount++;
                }
            }
            catch (ManagementException)
            {
                return -1;
            }
            return -1;
        }

        /// <summary> 
        /// Returns the total number of tracks on this hard drive 
        /// </summary> 
        /// <param name="drive"> 
        /// The drive to be parsed 
        /// </param> 
        /// <returns> 
        /// int: the number of tracks on the drive 
        /// </returns> 
        public static int GetTotalTracks(int drive)
        {
            int DriveCount = 0;
            try
            {
                ManagementObjectSearcher searcher =
                    new ManagementObjectSearcher("root\\CIMV2",
                    "SELECT * FROM Win32_DiskDrive");

                foreach (ManagementObject queryObj in searcher.Get())
                {
                    if (DriveCount == drive)
                    {
                        var t = queryObj["TotalTracks"];
                        return int.Parse((t.ToString()));
                    }
                    DriveCount++;
                }

            }
            catch (ManagementException)
            {
                return -1;
            }
            return -1;
        }

        /// <summary> 
        /// Returns the Sector from the drive at the specified location 
        /// </summary> 
        /// <param name="drive"> 
        /// The drive to have a sector read 
        /// </param> 
        /// <param name="sector"> 
        /// The sector number to read. 
        /// </param> 
        /// <param name="bytesPerSector"></param> 
        /// <returns></returns> 
        internal byte[] DumpSector(string drive, double sector, int bytesPerSector)
        {
            //short FILE_ATTRIBUTE_NORMAL = 0x80;
            //short INVALID_HANDLE_VALUE = -1;
            uint GENERIC_READ = 0x80000000;
            //uint GENERIC_WRITE = 0x40000000;
            //uint CREATE_NEW = 1;
            //uint CREATE_ALWAYS = 2;
            uint OPEN_EXISTING = 3;
            uint FILE_SHARE_READWRITE = 0x01 | 0x02;

            SafeFileHandle handleValue = UnsafeNativeMethods.CreateFile(drive, GENERIC_READ, FILE_SHARE_READWRITE, IntPtr.Zero, OPEN_EXISTING, 0, IntPtr.Zero);
            if (handleValue.IsInvalid)
            {
                Marshal.ThrowExceptionForHR(Marshal.GetHRForLastWin32Error());
            }
            double sec = sector * bytesPerSector;

            int size = int.Parse(bytesPerSector.ToString());
            byte[] buf = new byte[size];
            int read = 0;
            int moveToHigh;
            UnsafeNativeMethods.SetFilePointer(handleValue, int.Parse(sec.ToString()), out moveToHigh, EMoveMethod.Begin);
            ReadFile(handleValue, buf, size, out read, IntPtr.Zero);
            handleValue.Close();
            return buf;
        }

        /// <summary> 
        /// Returns the Sector from the drive at the specified location 
        /// </summary> 
        /// <param name="drive"> 
        /// The drive to have a sector read 
        /// </param> 
        /// <param name="sector"> 
        /// The sector number to read. 
        /// </param> 
        /// <param name="bytesPerSector"></param> 
        /// <returns></returns> 
        internal byte[] DumpTrack(string drive, double track, int bytesPerTrack, int TrackBufferSize)
        {
            //short FILE_ATTRIBUTE_NORMAL = 0x80;
            //short INVALID_HANDLE_VALUE = -1;
            uint GENERIC_READ = 0x80000000;
            //uint GENERIC_WRITE = 0x40000000;
            //uint CREATE_NEW = 1;
            //uint CREATE_ALWAYS = 2;
            uint OPEN_EXISTING = 3;



            SafeFileHandle handleValue = CreateFile(drive, GENERIC_READ, 0, IntPtr.Zero, OPEN_EXISTING, 0, IntPtr.Zero);
            if (handleValue.IsInvalid)
            {
                Marshal.ThrowExceptionForHR(Marshal.GetHRForLastWin32Error());
            }
            double trx = (track * bytesPerTrack * TrackBufferSize);

            int size = int.Parse(bytesPerTrack.ToString());
            byte[] buf = new byte[size * TrackBufferSize];
            int read = 0;
            int moveToHigh;
            SetFilePointer(handleValue, int.Parse(trx.ToString()), out moveToHigh, EMoveMethod.Begin);
            ReadFile(handleValue, buf, size, out read, IntPtr.Zero);
            handleValue.Close();
            return buf;
        }
    }
}
